/***************************************************************************
 *   Copyright (C) 2007 by GNB	(Grupo Computacion Biologica), EPS (Madrid)*
 *   Carlos Muñiz Solaz							   * 
 *   carlos.muniz@uam.es   						   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "motor.h"


#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
MODULE_AUTHOR("GNB-UAM");
MODULE_DESCRIPTION("Real-Time Task for controlling a McLennan P535 Stepper Motor");
#endif

#define DEBUG_LOG(x) rt_printk (x);

// In NI-PCIMIO 16, subdevice number 2 is used for commanding the motor
#define AO_SUBDEV_NUMBER 2


#define N_COMMANDS_PER_REV 2400

EXPORT_NO_SYMBOLS;

// Global symbols
static RT_TASK     rt_thread_motor;	// RT Task
comedi_t *devMotor;        		// Comedi device
MOTOR_SHM *motor_shm;			// Shared memory between the RT task and the GUI


unsigned int leftstep[] = { 10, 6, 5, 9 };
unsigned int rightstep[] =  {  9, 5, 6, 10 };
unsigned int mask = 0x000F;


// Keeps the track of the current step
int step;


/////////////////////////////////////// Initializing ///////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////
// Initializes motor and obtains module resources				
/////////////////////////////////////////////////////////////////////////////////
int motor_init(void)
{
    DEBUG_LOG ("Initializing motor ... \n");

    // Initializing shared memory
    if ( init_shmem() )
        goto err_shm;

    // Initializing comedi device
    if ( init_comedi() )
        goto err_comedi;

    // Starts real-time task
    if ( init_rt_thread() )
        goto error_thread;


    DEBUG_LOG ("Motor correctly initialized \n");


    return 0;

error_thread:
    cleanup_rt_thread();
err_comedi:
    cleanup_comedi();
err_shm:
    cleanup_shmem();

    DEBUG_LOG ("Could not initialize motor \n");

     // Could not initialize motor
    return 1;
}

/////////////////////////////////////////////////////////////////////////////////
// Initializes motor and obtains module resources				
/////////////////////////////////////////////////////////////////////////////////
int init_shmem ( void )
{
    DEBUG_LOG ("Initializing shared memory ...\n");

    // Creates the shared memory between the GUI and the module
    if (!(motor_shm = (MOTOR_SHM *) rtai_kmalloc(nam2num(MOTORSHM), sizeof(MOTOR_SHM))))
    {
        DEBUG_LOG ("ERROR: Failure to initialize shared memory.\n");
        return 1;
    }

    // Initializes the shared memory structure
    init_motor ( motor_shm );

    return 0;
}

/////////////////////////////////////////////////////////////////////////////////
// Initializes motor model structure				
/////////////////////////////////////////////////////////////////////////////////
void init_motor (MOTOR_SHM * motor_shm)
{
    motor_shm->speed = 30;
    motor_shm->right_angle = 10.0;
    motor_shm->left_angle = -10.0;
    motor_shm->n_steps = 0;
    motor_shm->direction = RIGHT;
    motor_shm->state_flag = CALIBRATE;
    motor_shm->curr_position = 0;
    motor_shm->bound_right = 45;
    motor_shm->bound_left = -45;
    motor_shm->T = calculatePeriod (motor_shm->speed );
    motor_shm->t_accel = 0.0;
    motor_shm->periodic = 0.0;
    motor_shm->incremental = 0.0;
    motor_shm->initial_t = 0.0;	    // Time when a periodic movement is started
    motor_shm->final_t = 0.0;	  // Time when a periodic movement should stop (period time)
    motor_shm->t = 0.0;

    // Motor model, for activity dependent experiments
    motor_shm->active_model = 0;
    motor_shm->threshold = 10.0;
    motor_shm->rate = 10;
    motor_shm->spike_ref_period = 1.0;
    motor_shm->burst_ref_period = 2000.0;
    motor_shm->burst_window = 20000.0;
    motor_shm->protocol = 0;
    motor_shm->init_direction = 0;
}

/////////////////////////////////////////////////////////////////////////////////
// Initializes motor model structure				
/////////////////////////////////////////////////////////////////////////////////
int init_comedi(void)
{
    int ret;

    DEBUG_LOG ("Initializing Comedi Device ...\n");

    // open device
    if (!(devMotor = comedi_open("/dev/comedi0")))
    {
        printk("ERROR: fail to open device /dev/comedi0\n");
        return 1;
    }

    // COnfigure device
    ret = comedi_dio_config (devMotor, AO_SUBDEV_NUMBER, 0, COMEDI_OUTPUT);
    ret = comedi_dio_config (devMotor, AO_SUBDEV_NUMBER, 1, COMEDI_OUTPUT);
    ret = comedi_dio_config (devMotor, AO_SUBDEV_NUMBER, 2, COMEDI_OUTPUT);
    ret = comedi_dio_config (devMotor, AO_SUBDEV_NUMBER, 3, COMEDI_OUTPUT);


    // Current step reset to zero
    step = 0;

    return 0;
}

/////////////////////////////////////////////////////////////////////////////////
// Initializes and launches RT task				
/////////////////////////////////////////////////////////////////////////////////
int init_rt_thread(void)
{
    // Create real time task
    if ( rt_task_init(&rt_thread_motor, RTtask, 0, 3000, 0, 1, 0) != 0)
    {
        rt_printk("ERROR: Failure to create RT Thread.\n");
        return 1;
    }

    // Launch real time task
    RTIME period = start_rt_timer(nano2count(motor_shm->T));
    rt_task_make_periodic(&rt_thread_motor, rt_get_time(), period);

    return 0;
}
/////////////////////////////////////// End of Initializing ////////////////////////////////////////////


/////////////////////////////////////// Real-Time Task ////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////
// Periodic RT task
/////////////////////////////////////////////////////////////////////////////////
void RTtask(int p)
{
    while (1)
    {

        switch (motor_shm->state_flag)
        {
            case CHANGE_SPEED:
                change_speed();
                break;
            case LEFT_MOV:
                move_left();
                break;
            case RIGHT_MOV:
                move_right();
                break;
            case RESET_MOTOR:
                reset_motor();
                break;
            case CALIBRATE:
                zeroing();
                break;
            case SET_PERIODIC:
                set_periodic();
                break;
            case FOREVER:
                rt_printk ("setting mode to forever periodic");
                motor_shm->forever = 1;
                motor_shm->state_flag = RIGHT_MOV; // We move in one direction arbitrarily left or right
                break;
            case INC_MODE:
                set_inc_mode();
                break;
            case ACCELERATE:
                accelerate();
                break;
            default:
                break;
        }

        motor_shm->t += motor_shm->T * 1e-6;

        rt_task_wait_period();
    } 

    return;
}// End of RTtask

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////// Helper Functions //////////////////////////////////////////////////

void reset_motor()
{
    DEBUG_LOG ("Resetting motor ... \n");
    motor_shm->periodic = 0;
    motor_shm->incremental = 0;
    motor_shm->forever = 0;
    motor_shm->accelerate = 0;

    if (motor_shm->n_steps < 0 )
    {
        motor_shm->right_angle = 0.0;
        motor_shm->state_flag = RIGHT_MOV;
    }
    else
    {
        motor_shm->left_angle = 0.0;
        motor_shm->state_flag = LEFT_MOV;
    }
}

void zeroing()
{
    DEBUG_LOG ("Zeroing ... \n");
    motor_shm->curr_position = 0;
    motor_shm->n_steps_right=0;
    motor_shm->n_steps_left=0;
    motor_shm->n_steps=0;
    motor_shm->state_flag = STOP_MOTOR;
}

void set_periodic()
{
    DEBUG_LOG ("Periodic Mode ... \n");
    motor_shm->periodic = 1;
    motor_shm->initial_t = motor_shm->t;
    motor_shm->state_flag = RIGHT_MOV; // We move in one direction arbitrarily left or right
}

void set_inc_mode() // Step by step
{
    DEBUG_LOG ("Incremental Mode ... \n");
    motor_shm->incremental = 1;
    motor_shm->right_angle = motor_shm->curr_position + STEP_ANGLE;
    motor_shm->left_angle = motor_shm->curr_position - STEP_ANGLE;
}

void change_speed()
{
    DEBUG_LOG ("Changing Speed ...\n");
    motor_shm->T = calculatePeriod (motor_shm->speed);
    rt_thread_motor.period = nano2count( motor_shm->T );
}

void accelerate()
{
    motor_shm->accelerate = 1;
    motor_shm->initial_t = motor_shm->t;

    motor_shm->T = calculateAperiod (motor_shm->speed);
    rt_thread_motor.period = nano2count( motor_shm->T );
    //rt_task_make_periodic(&rt_thread_motor,rt_get_time(),start_rt_timer(nano2count( motor_shm->T )));

    rt_printk ("\nSetting mode to acceleration");
    if (motor_shm->direction == RIGHT)
    {
        motor_shm->state_flag = RIGHT_MOV;
        return;
    }
    if (motor_shm->direction == LEFT)
    {
        motor_shm->state_flag = LEFT_MOV;
        return;
    }
}

double my_round (double x)
{
    if ((fabs(x) - (int) fabs(x)) > 0.5)
    {
        if (x < 0.0)
        {
            return (int) (x - 1.0);
        }
        else
        {
            return (int) (x + 1.0);
        }
    }

    return x;
}


void move_left()
{
    double diff;
    double diff2;
    double total;
    unsigned int data;


    rt_printk ("\nOscillating left\n");
    rt_printk ("\nmotor_shm->periodic: %ld\n", motor_shm->periodic);

    //  rt_printk ("t: %ld\nfinal_t: %ld", (int)(10000*motor_shm->t), (int)(10000*(motor_shm->initial_t + motor_shm->final_t)));

    if ( (motor_shm->periodic == 1) && (motor_shm->t > (motor_shm->initial_t + motor_shm->final_t)) )
    {
        rt_printk ("Stopping motor in left");
        motor_shm->state_flag = STOP_MOTOR;
        return;
    }


    if ( (motor_shm->accelerate == 1) && (motor_shm->t > (motor_shm->initial_t + motor_shm->t_accel)))
    {
        motor_shm->state_flag = STOP_MOTOR;
        rt_printk("stop time");
        return;
    }

    if ((motor_shm->curr_position <= ( motor_shm->bound_left )) || (motor_shm->curr_position <= ( motor_shm->left_angle)) )
    {
        if ( motor_shm->forever == 1)
        {
            motor_shm->state_flag = RIGHT_MOV;
            rt_printk("\n periodic movement changed from left to right");
            rt_printk ("\n angle left : %ld", (long int)(1000*motor_shm->left_angle));
            return;
        }
        if (motor_shm->accelerate == 1)
        {
            motor_shm->state_flag = STOP_MOTOR;
            rt_printk("\nStop angle");
            return;
        }
        if ( motor_shm->periodic == 1)  // Oscillate
        {
            motor_shm->state_flag = RIGHT_MOV;
            rt_printk("periodic movement changed from left to right");
            return;
        }
        else
        {
            motor_shm->state_flag = STOP_MOTOR;
            return;
        }
    }



    diff = fabs(my_round(motor_shm->left_angle - motor_shm->curr_position));
    diff2 = fabs(my_round(motor_shm->left_angle - motor_shm->curr_position + 0.15));

    if (diff < diff2)
        total = diff;
    else
        total = diff2 + 0.15;


    //          rt_printk ("left_angle: %ld %ld %ld %ld\n",  (int) (100.0*motor_shm->left_angle), (int) (100.0*motor_shm->curr_position), (int) (100.0*diff),(int) (100.0*total));

    if ((motor_shm->curr_position <= ( motor_shm->bound_left )) || (total <= 0.15) )
    {
        if ( motor_shm->forever == 1)
        {
            motor_shm->state_flag = RIGHT_MOV;
            rt_printk("\n periodic movement changed from left to right");
            rt_printk ("\n angle left : %ld", (long int)(1000*motor_shm->left_angle));
            return;
        }
        if (motor_shm->accelerate == 1)
        {
            motor_shm->state_flag = STOP_MOTOR;
            rt_printk("\nStop angle");
            return;
        }
        if ( motor_shm->periodic == 1)  // Oscillate
        {
            motor_shm->state_flag = RIGHT_MOV;
            rt_printk("periodic movement changed from left to right");
            return;
        }
        else
        {
            motor_shm->state_flag = STOP_MOTOR;
            return;
        }
    }
    else
    {

        data =leftstep[ step ];
        comedi_dio_bitfield (devMotor, AO_SUBDEV_NUMBER, mask, &data);
        step = (step+1)%4;
        motor_shm->curr_position -= STEP_ANGLE;
        //motor_shm->n_steps_left++;
        motor_shm->n_steps--;
        //     rt_printk ("n_steps: %ld\n", (int)motor_shm->n_steps);
        
    }
}

void move_right()
{
    double diff;
    double diff2;
    double total;
    unsigned int data;


    rt_printk ("\nOscillating right\n");
    rt_printk ("\nmotor_shm->periodic: %ld\n", motor_shm->periodic);

    /*rt_printk ("time: %ld\nfinal_t: %ld", (int)(10000*motor_shm->t), (int)(10000*(motor_shm->initial_t + motor_shm->t_accel)));
    rt_printk ("\nperiod accel : %ld", (long int)(1000*motor_shm->T));
    rt_printk ("min step %ld", (long int)(1000*calculatePeriod(MIN_SPEED)));*/

    rt_printk ("time: %ld\nfinal_t: %ld", (int)(10000*motor_shm->t), (int)(10000*(motor_shm->initial_t + motor_shm->final_t)));
    
    if ( (motor_shm->periodic == 1) && (motor_shm->t > (motor_shm->initial_t + motor_shm->final_t)))
    {
        rt_printk ("Stopping motor in right");
        motor_shm->state_flag = STOP_MOTOR;
        return;
    }

    if ( (motor_shm->accelerate == 1) && (motor_shm->t > (motor_shm->initial_t + motor_shm->t_accel)))
    {
        motor_shm->state_flag = STOP_MOTOR;
        rt_printk("stop time");
        return;
    }

    diff =fabs(my_round(motor_shm->right_angle - motor_shm->curr_position));
    diff2 = fabs(my_round(motor_shm->right_angle - motor_shm->curr_position - 0.15));

    if (diff < diff2)
        total = diff;
    else
        total = diff2 + 0.15;

//     rt_printk ("right_angle: %ld %ld %ld %ld\n",  (int) (100.0*motor_shm->right_angle), (int) (100.0*motor_shm->curr_position), (int) (100.0*diff),(int) (100.0*diff2));

    if ((motor_shm->curr_position >= motor_shm->bound_right) || (total < 0.15))
    {
        rt_printk ("bound_right: %ld, right_angle: %ld", (long int)motor_shm->bound_right,  (long int) motor_shm->right_angle);
        if (motor_shm->accelerate == 1)
        {
            motor_shm->state_flag = STOP_MOTOR;
            rt_printk("\nStop angle");
            return;
        }
        if ( motor_shm->forever == 1)
        {
            motor_shm->state_flag = LEFT_MOV;
            rt_printk("periodic movement changed from right to left");
            return;
        }
        if ( motor_shm->periodic == 1)  // Oscillate
        {
            motor_shm->state_flag = LEFT_MOV;
            rt_printk("periodic movement changed from right to left");
            return;
        }
        else
        {
            motor_shm->state_flag = STOP_MOTOR;
            return;
        }
    }
    else
    {
        data = rightstep[ step ];
        comedi_dio_bitfield (devMotor, AO_SUBDEV_NUMBER, mask, &data);
        step = (step+1)%4;
        motor_shm->curr_position += STEP_ANGLE;
        //  motor_shm->n_steps_right++;
        motor_shm->n_steps++;
        // rt_printk ("n_steps: %ld\n", (int)motor_shm->n_steps);
    }

}


double calculatePeriod (double speed)
{
    double f, T;

    if (speed < MIN_SPEED )
        speed = MIN_SPEED;

    if (speed > MAX_SPEED )
        speed = MAX_SPEED;

    f= (N_COMMANDS_PER_REV*speed)/360;

    if ( f > 200 )
    {
        f = 200;
    }

    T = (1.0/f)*1e3*1e6; //en ns
    return T;
}

double calculateAperiod(double speed)
{
    double t_step = 0.0;
    double speed1 = 0.0;

    if ( motor_shm->acceleration >=0  )
    {
        if (speed >= MAX_SPEED)
        {
            speed = MAX_SPEED;
            t_step = 1/((N_COMMANDS_PER_REV*speed)/360);
            t_step = t_step*1e9;
            motor_shm->speed = speed;
        }
        else
        {
            t_step = (-speed + sqrt(speed*speed + 2*motor_shm->acceleration*STEP_ANGLE))/(motor_shm->acceleration); //t_step (s)
            t_step = t_step*1e9; // t_step (ns)
            speed = sqrt(speed*speed + 2*motor_shm->acceleration*STEP_ANGLE);
            motor_shm->speed = speed;
        }
    }

    else
    {
        if  (speed <= MIN_SPEED)
        {
            speed = MIN_SPEED;
            t_step = 1/((N_COMMANDS_PER_REV*speed)/360);
            t_step = t_step*1e9;
            motor_shm->speed = speed;
        }
        else
        {
            t_step = (-speed + sqrt(speed*speed + 2*motor_shm->acceleration*STEP_ANGLE))/(motor_shm->acceleration); //t_step (s)
            t_step = t_step*1e9; // t_step (ns)
            speed1 = speed*speed + 2*motor_shm->acceleration*STEP_ANGLE;

            if (speed1 < 0)
            {
                speed1 = MIN_SPEED;
                motor_shm->speed = speed1;
            }
            else
            {
                speed1 = sqrt(speed1);
                motor_shm->speed = speed1;
            }
        }
    }
    return t_step;
}



///////////////////////////////////////////////// Releasing resources ///////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////
// Cleanup resources before unloading module				
/////////////////////////////////////////////////////////////////////////////////
void motor_exit(void)
{
    DEBUG_LOG ("Exiting motor module ...");
    cleanup_shmem();
    cleanup_comedi();
    cleanup_rt_thread();
    DEBUG_LOG ("Motor module exited correctly ...");
}

/////////////////////////////////////////////////////////////////////////////////
// Cleanup shared memory			
/////////////////////////////////////////////////////////////////////////////////
void cleanup_shmem( void )
{
    DEBUG_LOG ("Cleaning up Shared Memory \n");

    // Free reserved shared memory
    rtai_kfree(nam2num(MOTORSHM));
}

/////////////////////////////////////////////////////////////////////////////////
// Cleanup Comedi library		
/////////////////////////////////////////////////////////////////////////////////
void cleanup_comedi(void)
{
    DEBUG_LOG ("Cleaning up comedi device \n");

    // Free comedi device
    comedi_cancel(devMotor,AO_SUBDEV_NUMBER);
    comedi_close(devMotor);
    return;
}

/////////////////////////////////////////////////////////////////////////////////
// Cleanup the RT task		
/////////////////////////////////////////////////////////////////////////////////
void cleanup_rt_thread(void)
{
    DEBUG_LOG ("Cleaning up RT task \n");

    // Deletes RT task
    rt_task_delete(&rt_thread_motor);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

module_init(motor_init);
module_exit(motor_exit);
